#include <iostream>
#include <unistd.h>
#include "port.h"
//#include <malloc.h>

#include <ion.h>
#include <ion/events.h>
#include <ion/keyboard.h>
#include <omg/global_box.h>
#include <poincare/preferences.h>
#include "../../apps/apps_container.h"
#include "../../apps/global_preferences.h"
#include <kandinsky/ion_context.h>

#include <math.h>
#include <setjmp.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

/* py/parsenum.h is a C header which uses C keyword restrict.
 * It does not exist in C++ so we define it here in order to be able to include
 * py/parsenum.h header. */
#ifdef __cplusplus
#define restrict  // disable
#endif

extern "C" {
#include "mod/matplotlib/pyplot/modpyplot.h"
#include "mod/turtle/modturtle.h"
#include "mphalport.h"
#include "py/builtin.h"
#include "py/compile.h"
#include "py/gc.h"
#include "py/lexer.h"
#include "py/mperrno.h"
#include "py/mphal.h"
#include "py/nlr.h"
#include "py/parsenum.h"
#include "py/repl.h"
#include "py/runtime.h"
#include "py/stackctrl.h"
}

#include <escher/palette.h>

#if 1 // def GIAC_NUMWORKS
extern "C" const char * caseval(const char *);
#ifndef DEVICE
extern "C" const char * flash_read(const char *);
extern "C" int flash_filebrowser(const char ** filenames,int maxrecords,const char * extension);
#endif
#endif

static MicroPython::ScriptProvider *sScriptProvider = nullptr;
static MicroPython::ExecutionEnvironment *sCurrentExecutionEnvironment =
    nullptr;

MicroPython::ExecutionEnvironment::~ExecutionEnvironment() {
  sCurrentExecutionEnvironment = nullptr;
}

MicroPython::ExecutionEnvironment *
MicroPython::ExecutionEnvironment::currentExecutionEnvironment() {
  return sCurrentExecutionEnvironment;
}

bool MicroPython::ExecutionEnvironment::runCode(const char *str) {
  assert(sCurrentExecutionEnvironment == nullptr);
  sCurrentExecutionEnvironment = this;

  /* Set the user interruption now, as it is needed for the normal execution and
   * for the exception handling (because of print). */
  mp_hal_set_interrupt_char((int)Ion::Keyboard::Key::Back);

  bool runSucceeded = true;
  nlr_buf_t nlr;
  if (nlr_push(&nlr) == 0) {
    mp_lexer_t *lex = mp_lexer_new_from_str_len(0, str, strlen(str), false);
    /* The input type is "single input" because the Python console is supposed
     * to be fed lines and not files. */
    // TODO: add a parameter when other input types (file, eval) are required
    mp_parse_tree_t pt = mp_parse(lex, MP_PARSE_SINGLE_INPUT);
    mp_obj_t module_fun = mp_compile(&pt, lex->source_name, true);
    mp_call_function_0(module_fun);
    nlr_pop();
  } else {  // Uncaught exception
    runSucceeded = false;
    HandleException(&nlr);
  }

  // Disable the user interruption
  mp_hal_set_interrupt_char(-1);

  assert(sCurrentExecutionEnvironment == this);
  sCurrentExecutionEnvironment = nullptr;
  return runSucceeded;
}

void MicroPython::ExecutionEnvironment::HandleExceptionSilently() {
  // Flush the store if an error is encountered to avoid being stuck with a full
  // memory
  modpyplot_flush_used_heap();
  // TODO: do the same for other modules?
}

void MicroPython::ExecutionEnvironment::HandleException(
    nlr_buf_t *nlr_buf, MicroPython::ExecutionEnvironment *env) {
  /* We need a static execution environment to print. If it's not already set,
   * the caller should give it as a parameter. */
  if (env != nullptr) {
    assert(sCurrentExecutionEnvironment == nullptr);
    sCurrentExecutionEnvironment = env;
  }
  assert(sCurrentExecutionEnvironment != nullptr);

  /* If the exception happens while displaying the view controller, we need to
   * hide it to avoid redrawing uncomplete objects. */
  sCurrentExecutionEnvironment->hideAnyDisplayedViewController();

  /* mp_obj_print_exception is supposed to handle error printing. However,
   * because we want to print custom information, we copied and modified the
   * content of mp_obj_print_exception instead of calling it. */
  if (mp_obj_is_exception_instance((mp_obj_t)(nlr_buf->ret_val))) {
    size_t n, *values;
    mp_obj_exception_get_traceback((mp_obj_t)(nlr_buf->ret_val), &n, &values);
    if (n > 0) {
      assert(n % 3 == 0);
      for (int i = n - 3; i >= 0; i -= 3) {
        if (values[i] != 0 || i == 0) {
          if (values[i] == 0) {
            mp_printf(&mp_plat_print, "  Last command\n");
          } else {
#if MICROPY_ENABLE_SOURCE_LINE
            mp_printf(&mp_plat_print, "  File \"%q\", line %d", values[i],
                      (int)values[i + 1]);
#else
            mp_printf(&mp_plat_print, "  File \"%q\"", values[i]);
#endif
            // the block name can be NULL if it's unknown
            qstr block = values[i + 2];
            if (block == MP_QSTRnull) {
              mp_print_str(&mp_plat_print, "\n");
            } else {
              mp_printf(&mp_plat_print, ", in %q\n", block);
            }
          }
        }
      }
    }
  }
  mp_obj_print_helper(&mp_plat_print, (mp_obj_t)(nlr_buf->ret_val), PRINT_EXC);
  mp_print_str(&mp_plat_print, "\n");
  /* End of mp_obj_print_exception. */

  HandleExceptionSilently();

  // Clean the static ExecutionEnvironment
  if (env != nullptr) {
    sCurrentExecutionEnvironment = nullptr;
  }
}

void MicroPython::ExecutionEnvironment::interrupt() { mp_keyboard_interrupt(); }

extern "C" {
extern const void *_process_stack_end;
}

void MicroPython::init(void *heapStart, void *heapEnd) {
#if __EMSCRIPTEN__
  static mp_obj_t pystack[1024];
  mp_pystack_init(pystack, &pystack[MP_ARRAY_SIZE(pystack)]);
#endif
  /* We delimit the stack part that will be used by Python. The stackTop is the
   * address of the first object that can be allocated on Python stack. This
   * boundaries are used:
   * - by gc_collect to determine where to collect roots of the objects that
   *   must be kept on the heap
   * - to check if the maximal recursion depth has been reached.
   * Current stack pointer could go backward after initialization. A stack start
   * pointer defined in main is therefore used. */
  void *stackTopAddress = Ion::stackStart();

#if MP_PORT_USE_STACK_SYMBOLS
  mp_stack_set_top(stackTopAddress);
  size_t stackLimitInBytes =
      (char *)stackTopAddress - (char *)&_process_stack_end;
  mp_stack_set_limit(stackLimitInBytes);
#else
  mp_stack_set_top(stackTopAddress);
  /* The stack limit is set to roughly mimic the maximal recursion depth of the
   * device - and actually to be slightly less to be sure not to beat the device
   * performance.  */
  mp_stack_set_limit(29152);
#endif
  gc_init(heapStart, heapEnd);
  mp_init();
}

void MicroPython::deinit() { mp_deinit(); }

void MicroPython::registerScriptProvider(ScriptProvider *s) {
  sScriptProvider = s;
}

void MicroPython::collectRootsAtAddress(char *address, int byteLength) {
  /* The given address is not necessarily aligned on sizeof(void *). However,
   * any pointer stored in the range [address, address + byteLength] will be
   * aligned on sizeof(void *). This is a consequence of the alignment
   * requirements of compilers (Cf http://www.catb.org/esr/structure-packing/).
   * Micropython gc_collect_root scans looking for pointers jumping every
   * sizeof(void *). It has to be provided with a sizeof(uintptr_t)-aligned
   * address. */
  // Compute the aligned address
  // 0b000...00011 with 2 (or 3 for x64 arch) 1s
  uintptr_t bitMaskOnes = sizeof(uintptr_t) - 1;
  // 0b111...11100 with sizeof(uintptr_t)-1 0s
  uintptr_t bitMaskZeros = ~bitMaskOnes;
  uintptr_t alignedAddress =
      reinterpret_cast<uintptr_t>(address) & bitMaskZeros;
  /* Increase the length consequently with the new alignment
   * (We don't need to increase the byteLength to a sizeof(uintptr_t)-aligned
   * lenght because no pointer can be stored on less than sizeof(uintptr_t)
   * bytes.) */
  int alignedByteLength = byteLength;
  alignedByteLength += reinterpret_cast<uintptr_t>(address) & bitMaskOnes;

  assert(alignedAddress % ((uintptr_t)sizeof(uintptr_t)) == 0);
  gc_collect_root((void **)alignedAddress,
                  alignedByteLength / sizeof(uintptr_t));
}

KDColor MicroPython::Color::Parse(mp_obj_t input, Mode mode) {
  constexpr static int maxColorIntensity =
      static_cast<int>(Mode::MaxIntensity255);
  if (mp_obj_is_str(input)) {
    size_t l;
    const char *color = mp_obj_str_get_data(input, &l);
    constexpr NamedColor pairs[] = {
        NamedColor("blue", KDColorBlue),
        NamedColor("b", KDColorBlue),
        NamedColor("red", KDColorRed),
        NamedColor("r", KDColorRed),
        NamedColor("green", Escher::Palette::Green),
        NamedColor("g", Escher::Palette::Green),
        NamedColor("yellow", KDColorYellow),
        NamedColor("y", KDColorYellow),
        NamedColor("brown", Escher::Palette::Brown),
        NamedColor("black", KDColorBlack),
        NamedColor("k", KDColorBlack),
        NamedColor("white", KDColorWhite),
        NamedColor("w", KDColorWhite),
        NamedColor("pink", Escher::Palette::Pink),
        NamedColor("orange", Escher::Palette::Orange),
        NamedColor("purple", Escher::Palette::Purple),
        NamedColor("gray", Escher::Palette::GrayDark),
        NamedColor("grey", Escher::Palette::GrayDark),
        NamedColor("cyan", Escher::Palette::Cyan),
        NamedColor("magenta", Escher::Palette::Magenta)};
    for (NamedColor p : pairs) {
      if (strcmp(p.name(), color) == 0) {
        return p.color();
      }
    }

    if (color[0] == '#') {
      // TODO handle #abc as #aabbcc (see matplotlib spec)
      if (l != 7) {
        mp_raise_ValueError("RGB hex values are 6 bytes long");
      }
      uint32_t colorInt = mp_obj_get_int(
          mp_parse_num_integer(color + 1, strlen(color + 1), 16, NULL));
      return KDColor::RGB24(colorInt);
    }

    mp_float_t grayLevel = mp_obj_float_get(
        mp_parse_num_decimal(color, strlen(color), false, false, NULL));
    if (grayLevel >= 0.0 && grayLevel <= 1.0) {
      uint8_t color = maxColorIntensity * (float)grayLevel;
      return KDColor::RGB888(color, color, color);
    }
    mp_raise_ValueError("Gray levels are between 0.0 and 1.0");
  } else if (mp_obj_is_int(input)) {
    mp_raise_TypeError("Int are not colors");
    // See
    // https://github.com/numworks/epsilon/issues/1533#issuecomment-618443492
  } else {
    size_t len;
    mp_obj_t *elem;

    mp_obj_get_array(input, &len, &elem);

    if (len != 3) {
      mp_raise_TypeError("Color needs 3 components");
    }
    int intensityFactor = maxColorIntensity / static_cast<int>(mode);
    return KDColor::RGB888(intensityFactor * mp_obj_get_float(elem[0]),
                           intensityFactor * mp_obj_get_float(elem[1]),
                           intensityFactor * mp_obj_get_float(elem[2]));
  }
  mp_raise_TypeError("Color couldn't be parsed");
}

void gc_collect_regs_and_stack(void) {
  // get the registers and the sp
  jmp_buf regs;
  uintptr_t sp = Ion::collectRegisters(regs);

  void *python_stack_top = MP_STATE_THREAD(stack_top);
  assert(python_stack_top != NULL);

  /* On the device, the stack is stored in reverse order, but it might not be
   * the case on a computer. We thus have to take the absolute value of the
   * addresses difference. */
  size_t stackLengthInByte;
  void **scanStart;
  if ((uintptr_t)python_stack_top > sp) {
    /* To compute the stack length:
     *                               registers
     *                             <----------->
     * STACK <-  ...|  |  |  |  |  |--|--|--|--|  |  |  |  |  |  |
     *                             ^                             ^
     *                             sp                      python_stack_top
     * */

    stackLengthInByte = (uintptr_t)python_stack_top - sp;
    scanStart = (void **)sp;

  } else {
    /* When computing the stack length, take into account regs' size.
     *                                              registers
     *                                            <----------->
     * STACK ->  |  |  |  |  |  |  |  |  |  |  |  |--|--|--|--|  |  |  |...
     *           ^python_stack_top                ^sp
     * */

    stackLengthInByte = sp - (uintptr_t)python_stack_top + sizeof(regs);
    scanStart = (void **)python_stack_top;
  }
  /* Memory error detectors might find an error here as they might split regs
   * and stack memory zones. */
  MicroPython::collectRootsAtAddress((char *)scanStart, stackLengthInByte);
}

void gc_collect(void) {
  gc_collect_start();
  modturtle_gc_collect();
  modpyplot_gc_collect();
  gc_collect_regs_and_stack();
  gc_collect_end();
}

void nlr_jump_fail(void *val) {
  /* Called by nlr_jump if no nlr buf has been pushed.
   * Bail out with a fatal error. */
  abort();
}

extern "C" void console_output(const char *,int);
extern "C" const char * console_input(const char * msg1,const char * msg2,bool numeric,int ypos);

void mp_hal_stdout_tx_strn_cooked(const char *str, size_t len) {
  if (sCurrentExecutionEnvironment != nullptr)
    sCurrentExecutionEnvironment->printText(str, len);
  else
    console_output(str,len); 
}

const char *mp_hal_input(const char *prompt) {
  if (sCurrentExecutionEnvironment != nullptr)
    return sCurrentExecutionEnvironment->inputText(prompt);
  return strdup(console_input(prompt,"? ",false,185));
}

/* C-- SDK , (c) B. Parisse 2019 */
int parser_errorline=0,parser_errorcol=0,heap_size=32768;

int mp_token(const char * line){ // does not work, always mp_type_fun_bc
  mp_obj_dict_t *dict = NULL;
  dict = mp_locals_get();//mp_globals_get();//
  mp_map_t * m=mp_obj_dict_get_map(dict);
  mp_obj_t index=mp_obj_new_str(line,strlen(line));
  mp_map_elem_t *ptr=mp_map_lookup(m, index, MP_MAP_LOOKUP) ;
  // 0 or 3
  if (ptr)
    return 3;
  else
    return 0;
}

const char * const * mp_vars(){
  mp_obj_dict_t *dict = NULL;
  dict = mp_locals_get();

  mp_obj_t dir = mp_obj_new_list(0, NULL);
  if (dict != NULL) {
    for (size_t i = 0; i < dict->map.alloc; i++) {
      if (MP_MAP_SLOT_IS_FILLED(&dict->map, i)) {
	//const char * ch=mp_obj_str_get_str(dict->map.table[i].key);
	mp_obj_t v=dict->map.table[i].value;
	const mp_obj_type_t * vt=mp_obj_get_type(v);
	if (vt!=&mp_type_fun_builtin_var
	    && vt!=&mp_type_fun_builtin_0
	    && vt!=&mp_type_fun_builtin_1
	    && vt!=&mp_type_fun_builtin_2
	    && vt!=&mp_type_fun_builtin_3
	    )
	  mp_obj_list_append(dir, dict->map.table[i].key);
      }
    }
  }
  mp_obj_t * line; size_t m=0;
  mp_obj_get_array(dir,&m,&line);
  const char ** tab=(const char **)malloc((m+1)*sizeof(char *));
  if (!tab) return 0;
  for (size_t i=0;i<m;++i){
    tab[i]=mp_obj_str_get_str(line[i]);
  }
  tab[m]=0;
  for (size_t i=0;i<m;++i){
    if (strcmp(tab[i],"__name__")==0){
      tab[i]=tab[m-1];
      tab[m-1]="del ";
      break;
    }
  }
  return tab;
}

int micropy_eval(const char * str){
  nlr_buf_t nlr;
  if (nlr_push(&nlr) == 0) {
    mp_lexer_t *lex = mp_lexer_new_from_str_len(0, str, strlen(str), false);
    mp_parse_tree_t pt = mp_parse(lex, MP_PARSE_FILE_INPUT);
    mp_obj_t module_fun = mp_compile(&pt, lex->source_name, true);
    // mp_hal_set_interrupt_char((int)Ion::Keyboard::Key::Back);
    mp_call_function_0(module_fun);
    // mp_hal_set_interrupt_char(-1); // Disable interrupt
    nlr_pop();
  } else { // Uncaught exception
    /* mp_obj_print_exception is supposed to handle error printing. However,
     * because we want to print custom information, we copied and modified the
     * content of mp_obj_print_exception instead of calling it. */
    if (mp_obj_is_exception_instance((mp_obj_t)nlr.ret_val)) {
        size_t n, *values;
        mp_obj_exception_get_traceback((mp_obj_t)nlr.ret_val, &n, &values);
        if (n > 0) {
            assert(n % 3 == 0);
            for (int i = n - 3; i >= 0; i -= 3) {
              if (values[i] != 0 || i == 0) {
                if (values[i] == 0) {
                  mp_printf(&mp_plat_print, "  Last command\n");
                } else {
#if MICROPY_ENABLE_SOURCE_LINE
                  mp_printf(&mp_plat_print, "  File \"%q\", line %d", values[i], (int)values[i + 1]);
#else
                  mp_printf(&mp_plat_print, "  File \"%q\"", values[i]);
#endif
                  // the block name can be NULL if it's unknown
                  qstr block = values[i + 2];
                  if (block == MP_QSTRnull) {
                    mp_print_str(&mp_plat_print, "\n");
                  } else {
                    mp_printf(&mp_plat_print, ", in %q\n", block);
                  }
                }
              }
            }
        }
    }
    mp_obj_print_helper(&mp_plat_print, (mp_obj_t)nlr.ret_val, PRINT_EXC);
    mp_print_str(&mp_plat_print, "\n");
    /* End of mp_obj_print_exception. */
  }
  return 0;
}

char * micropy_init(int stack_size,int heap_size){
  mp_stack_set_limit(40000);

    //Disable output buffering, otherwise interactive mode becomes useless
    setbuf(stdout, NULL);
  
    mp_stack_set_limit(stack_size);// (32768);

#if MICROPY_ENABLE_GC
    char *heap = (char *)malloc(heap_size);
    if(!heap)
    {
	return 0;
    }
    gc_init(heap, heap + heap_size - 1);
#endif

    mp_init();

    return heap;
}


void enable_back_interrupt(){
  mp_interrupt_char = (int)Ion::Keyboard::Key::Back;
}

void disable_back_interrupt(){
  mp_interrupt_char = -1;
}

extern "C" const char * flash_read(const char * filename);

int filesize(const char * filename){
  const Ion::Storage::Record r=Ion::Storage::FileSystem::sharedFileSystem->recordNamed(filename);
  if (r.isNull()){
#ifdef DEVICE
    return 0;
#else
    return strlen(flash_read(filename));
#endif
  }
  Ion::Storage::Record::Data d=r.value();
  return d.size;
}

const char * read_file(const char * filename){
  const Ion::Storage::Record r=Ion::Storage::FileSystem::sharedFileSystem->recordNamed(filename);
  if (r.isNull()){
#ifdef DEVICE
    return 0;
#else
    return flash_read(filename);
#endif
  }
  Ion::Storage::Record::Data d=r.value();
  const char * ptr=(const char *)d.buffer;
  if (ptr)
    return ptr+1;
  else
    return 0;
}


bool write_file(const char * filename,const char * content,size_t len){
  const void * dataChunks[]={content};
  size_t sizeChunks[1]={len?len:strlen(content)+1};
  auto res=Ion::Storage::FileSystem::sharedFileSystem
    ->createRecordWithFullNameAndDataChunks(filename,dataChunks,sizeChunks,1);
  if (res==Ion::Storage::Record::ErrorStatus::None)
    return true;
  if (res==Ion::Storage::Record::ErrorStatus::NameTaken){
    auto r=Ion::Storage::FileSystem::sharedFileSystem->recordNamed(filename);
    Ion::Storage::Record::Data d;
    d.buffer=content;
    d.size=(len?len:strlen(content))+1;
    auto rr=r.setValue(d);
    bool b=rr==Ion::Storage::Record::ErrorStatus::None;
    return b;
  }
  return false;
}

bool file_exists(const char * filename){
  auto s=Ion::Storage::FileSystem::sharedFileSystem;
  auto r=s->recordNamed(filename);
  if (!r.isNull())
    return true;
  return read_file(filename);
}

mp_lexer_t *mp_lexer_new_from_file(const char *filename) {
#if 1
  const char * script=read_file(filename);
  if (script){
    return mp_lexer_new_from_str_len(qstr_from_str(filename), script,
                                     strlen(script), 0 /* size_t free_len*/);
  }
  else
    mp_raise_OSError(MP_ENOENT);
#else
  if (sScriptProvider != nullptr) {
    const char *script = sScriptProvider->contentOfScript(filename, true);
    if (script != nullptr) {
      return mp_lexer_new_from_str_len(qstr_from_str(filename), script,
                                       strlen(script), 0 /* size_t free_len*/);
    } else {
      mp_raise_OSError(MP_ENOENT);
    }
  } else {
    mp_raise_OSError(MP_ENOENT);
  }
#endif
}

mp_import_stat_t mp_import_stat(const char *path) {
  if (
      file_exists(path)
      //sScriptProvider && sScriptProvider->contentOfScript(path, false)
      ) {
    return MP_IMPORT_STAT_FILE;
  }
  return MP_IMPORT_STAT_NO_EXIST;
}

bool erase_file(const char * filename){
  auto s=Ion::Storage::FileSystem::sharedFileSystem;
  auto r= s->recordNamed(filename);
  if (r.isNull())
    return false;
  r.destroy();
  return true;
}


#if 1
void * ion_storage(){
  auto s=Ion::Storage::FileSystem::sharedFileSystem;
  return s;
}

int os_file_browser(const char ** filenames,int maxrecords,const char * extension,int storage){
  auto s=Ion::Storage::FileSystem::sharedFileSystem;
  int n=storage==1?0:s->numberOfRecordsWithExtension(extension?extension:"py");
  //if (!n) return 0; 
  if (n>maxrecords) n=maxrecords;
  for (int i=0;i<n;i++){
    const Ion::Storage::Record & r=s->recordWithExtensionAtIndex(extension?extension:"py", i);
    filenames[i]=r.fullName();
  }
  if (n==maxrecords || storage==0){
    filenames[n]=0;
    return n;
  }
  int flashn=flash_filebrowser(filenames+n,maxrecords-n,extension?extension:"py");
  return n+flashn;
}
  // const char * ptr=(const char *)r.value().buffer;
  /*     storage.h
	 Ion::Storage::Record::Data structure avec 2 membres
	 const void * buffer et size_t size

	 Record(const char * fullName = nullptr);
	 Record(const char * basename, const char * extension);
	 Data value();
	 Ion::Storage::Record::ErrorStatus setValue(Ion::Storage::Record::Data);
	 void destroy(void);

  // Record creation
  Record::ErrorStatus createRecordWithFullName(const char * fullName, const void * data, size_t size);
  Record::ErrorStatus createRecordWithExtension(const char * baseName, const char * extension, const void * data, size_t size);

  // Record getters
  Record recordWithExtensionAtIndex(const char * extension, int index);
  Record recordNamed(const char * fullName);
  Record recordBaseNamedWithExtension(const char * baseName, const char * extension);
  Record recordBaseNamedWithExtensions(const char * baseName, const char * const extension[], size_t numberOfExtensions);

  // Record destruction
  void destroyAllRecords();
  void destroyRecordWithBaseNameAndExtension(const char * baseName, const char * extension);
  void destroyRecordsWithExtension(const char * extension);


  */
#endif

void statuslinemsg(const char * msg){
  AppsContainer::sharedAppsContainer()->reloadTitleBarView();
  auto ptr=MicroPython::ExecutionEnvironment::currentExecutionEnvironment();
  if (ptr)
    ptr->displaySandbox();
  auto ctx=KDIonContext::SharedContext;
  KDRect save=ctx->m_clippingRect;
  KDPoint o=ctx->m_origin;
  ctx->setClippingRect(KDRect(0,0,280,18));
  ctx->setOrigin(KDPoint(0,0));
  KDRect rect(0,0,280,18);
  ctx->pushRectUniform(rect,64934 /* Palette::YellowDark*/);
  bool large_font=strlen(msg)>25;
  ctx->drawString(msg,KDPoint(0,0),
                  {.glyphColor = 0,
                   .backgroundColor = 64934,
                   .font = large_font ? KDFont::Size::Large : KDFont::Size::Small},
                  255);

  ctx->setClippingRect(save);
  ctx->setOrigin(o);
}

bool os_set_angle_unit(int mode){
  Poincare::Preferences * preferences = Poincare::Preferences::sharedPreferences;
 if (mode==0){ // rad
    preferences->setAngleUnit(Poincare::Preferences::AngleUnit::Radian);
    return true;
  }
  if (mode==1){ // deg
    preferences->setAngleUnit(Poincare::Preferences::AngleUnit::Degree);
    return true;
  }
  return false;
  if (mode==2){ // grad
    return true;
  }
}

int os_get_angle_unit(){
  Poincare::Preferences * preferences = Poincare::Preferences::sharedPreferences;
  if (preferences->angleUnit()==Poincare::Preferences::AngleUnit::Radian)
    return 0;
  if (preferences->angleUnit()==Poincare::Preferences::AngleUnit::Degree)
    return 1;
  return 2;
}

#ifdef SIMULATOR
#define TICKS_PER_MINUTE 60000
#else
#define TICKS_PER_MINUTE 11862
extern const void * _stack_start;
extern const void * _heap_start;
#endif
extern OMG::GlobalBox<Ion::Keyboard::State> SharedState;
bool khicas_isshifted;
//int persistance_restart=0;
//int time_shift=0; // set it via time() command in KhiCAS
void statusline(int mode){
  AppsContainer::sharedAppsContainer()->reloadTitleBarView();
  auto ctx=KDIonContext::SharedContext;
  KDRect save=ctx->m_clippingRect;
  KDPoint o=ctx->m_origin;
  ctx->setClippingRect(KDRect(0,0,320,18));
  ctx->setOrigin(KDPoint(0,0));
  KDRect rect(0,0,mode==1?320:280,18);
  KDIonContext::SharedContext->pushRectUniform(rect, 1855/* 64934 Palette::YellowDark*/);
  const char * text=0;
  Poincare::Preferences * preferences = Poincare::Preferences::sharedPreferences;
  if (preferences->angleUnit() == Poincare::Preferences::AngleUnit::Radian)
    text="rad";
  else
    text="deg";
  // ctx->drawString(text, KDPoint(5,1), KDFont::Font(KDFont::Size::Small), 0, 63488 /* Palette::Red*/);
  ctx->drawString(text, KDPoint(5,1),
      {.glyphColor = 0,
       .backgroundColor = 1855,
       .font = KDFont::Size::Small},
      255);
#ifdef GIAC_LINKED
  if (khicas_eval)
    text="KHICAS";
  else
    text="PYTHON";
#else
  text="KHICAS";
#endif
  ctx->drawString(text, KDPoint(70,1),
      {.glyphColor = 0,
       .backgroundColor = 1855,
       .font = KDFont::Size::Small},
      255);
  char bufheap[16];
  // malloc_trim(0); // not defined in newlib (nano specs?)
#if 1 // mingw64 does not support sbrk?
  size_t heap=0;
#else
  size_t heap=(size_t) sbrk(0);
#endif
  //heap=persistance_restart;
#if 0 //ndef SIMULATOR
  heap=(size_t)_stack_start - heap;
  // heap=heap-(size_t)_heap_start;
#endif
  int x;
  x=(heap&0xf0000000)>>28;
  bufheap[0]=(x>9?'a'+(x-10):'0'+x);
  x=(heap&0xf000000)>>24;
  bufheap[1]=(x>9?'a'+(x-10):'0'+x);
  x=(heap&0xf00000)>>20;
  bufheap[2]=(x>9?'a'+(x-10):'0'+x);
  x=(heap&0xf0000)>>16;
  bufheap[3]=(x>9?'a'+(x-10):'0'+x);
  x=(heap&0xf000)>>12;
  bufheap[4]=(x>9?'a'+(x-10):'0'+x);
  x=(heap&0xf00)>>8;
  bufheap[5]=(x>9?'a'+(x-10):'0'+x);
  x=(heap&0xf0)>>4;
  bufheap[6]=(x>9?'a'+(x-10):'0'+x);
  x=(heap&0xf);
  bufheap[7]=(x>9?'a'+(x-10):'0'+x);
  bufheap[8]=0;
  ctx->drawString(bufheap,KDPoint(130,1),
      {.glyphColor = 0,
       .backgroundColor = 1855,
       .font = KDFont::Size::Small},
      255);
#ifdef GIAC_SHOWTIME
  int d=(Ion::Timing::millis()/TICKS_PER_MINUTE +time_shift) % (24*60); // minutes
  char buf[32]={0,0,0,0};
  int h=d/60;
  buf[0]='0'+h/10;
  buf[1]='0'+(h%10);
  buf[2]=':';
  ctx->drawString(buf, KDPoint(148,1), KDFont::Font(KDFont::Size::Small), 0, 1855);
  int mn=d%60;
  buf[0]='0'+mn/10;
  buf[1]='0'+(mn%10);
  buf[2]=0;
  ctx->drawString(buf, KDPoint(168,1), KDFont::Font(KDFont::Size::Small), 0, 1855);
#endif
  text="     ";
  khicas_isshifted=Ion::Events::shiftAlphaStatus().shiftIsActive();
  if (isalphaactive()){
    if (khicas_isshifted){
      if (Ion::Events::shiftAlphaStatus().alphaIsLocked())
        text="ALPHAl";
      else
        text="1ALPHA";
    } else {
      if (Ion::Events::shiftAlphaStatus().alphaIsLocked())
        text="alphal";
      else
        text="1alpha";
    }
  }
  else {
    if (khicas_isshifted)
      text="shift ";
  }
  ctx->drawString(text, KDPoint(232,1),
      {.glyphColor = 0,
       .backgroundColor = 1855,
       .font = KDFont::Size::Small},
      255);
  if (mode==1){
    if (Ion::USB::isPlugged())
      text="charge";
    else {
      auto c=Ion::Battery::level();
      if (c==Ion::Battery::Charge::EMPTY)
	text="empty ";
      if (c==Ion::Battery::Charge::LOW)
	text="low   ";
      if (c==Ion::Battery::Charge::FULL)
	text="full  ";
    }
  ctx->drawString(text, KDPoint(280,1), 
      {.glyphColor = 0,
       .backgroundColor = 1855,
       .font = KDFont::Size::Small},
      255);
  }
  ctx->setClippingRect(save);
  ctx->setOrigin(o);
}

bool isalphaactive(){
  return Ion::Events::shiftAlphaStatus().alphaIsActive();
}

void lock_alpha(){
  Ion::Events::setShiftAlphaStatus(Ion::Events::ShiftAlphaStatus(Ion::Events::ShiftAlphaStatus::ShiftStatus(false),Ion::Events::ShiftAlphaStatus::AlphaStatus(2)));
  statusline(0);
}

void reset_kbd(){
  Ion::Events::setShiftAlphaStatus(Ion::Events::ShiftAlphaStatus(Ion::Events::ShiftAlphaStatus::ShiftStatus(false),Ion::Events::ShiftAlphaStatus::AlphaStatus(0)));
  statusline(0);
}

void os_redraw(){
  AppsContainer::sharedAppsContainer()->m_window.redraw(true);
}

double millis(){
  return double(Ion::Timing::millis()*60000.0/TICKS_PER_MINUTE);
}

bool alphawasactive_=false;
bool alphawasactive(){
  return alphawasactive_;
}

bool waitforvblank(){
  return Ion::Display::waitForVBlank();
}

bool back_key_pressed() {
  static int c = 0;

  ++c ;
  if (c<400 || (c & 0xf)!= 0) {
    return false;
  }

  Ion::Keyboard::State scan = Ion::Keyboard::scan();
  // if (scan!=16) std::cerr << scan << '\n';
  Ion::Keyboard::Key interruptKey = static_cast<Ion::Keyboard::Key>(mp_interrupt_char);
  if (scan.keyDown(interruptKey)
      // || scan.keyDown(static_cast<Ion::Keyboard::Key>(16))
      ) 
    return true;
  return false;
}

int getkey_raw(bool allow_suspend){
  int key=-1;
  size_t t1=Ion::Timing::millis();
  for (;;){
    int timeout=10000;
    alphawasactive_=isalphaactive();
    Ion::Events::Event event=Ion::Events::getEvent(&timeout);
    auto ctx=KDIonContext::SharedContext;
    KDRect save=ctx->m_clippingRect;
    KDPoint o=ctx->m_origin;
    ctx->setClippingRect(KDRect(0,0,320,240));
    ctx->setOrigin(KDPoint(0,18));
    KDRect rect(90,63,140,75);
    if (event==Ion::Events::None){
      size_t t2=Ion::Timing::millis();
      if (t2-t1>2*TICKS_PER_MINUTE){
	// ctx->pushRectUniform(rect,1855 /* Palette::YellowDark*/);
	event=Ion::Events::OnOff;
      }
    }
    else
      t1=Ion::Timing::millis();
    if (event.isKeyboardEvent()) {
      // m_backlightDimmingTimer.reset();
      // m_suspendTimer.reset();
      Ion::Backlight::setBrightness(GlobalPreferences::sharedGlobalPreferences->brightnessLevel());
    }
    ctx->setClippingRect(save);
    ctx->setOrigin(o);

    if (event==Ion::Events::Shift || event==Ion::Events::Alpha){
      statusline();
      continue;
    }
    if (event.isKeyboardEvent()){
      key=event.m_id; // static_cast<unsigned char>( event )
      if (khicas_isshifted && (key%108)/54==0)
        key += 54;
      if (key==17 || key==4 || key==5 || key==52)
	reset_kbd();
      if (allow_suspend && (key==7 || key==8) ){ // power
	Ion::Power::suspend(true);
	numworks_fill_rect(0,0,320,240,65535);
	Ion::Backlight::setBrightness(GlobalPreferences::sharedGlobalPreferences->brightnessLevel());
	AppsContainer::sharedAppsContainer()->reloadTitleBarView();
	//AppsContainer::sharedAppsContainer()->redrawWindow();
	statusline(1);
	//continue;
      }
      else
	statusline();
      break;
    }
  }
  return key;
}

const short int translated_keys[]=
  {
   // non shifted
   KEY_CTRL_LEFT,KEY_CTRL_UP,KEY_CTRL_DOWN,KEY_CTRL_RIGHT,KEY_CTRL_OK,KEY_CTRL_EXIT,
   KEY_CTRL_MENU,KEY_PRGM_ACON,KEY_PRGM_ACON,9,10,11,
   KEY_CTRL_SHIFT,KEY_CTRL_ALPHA,KEY_CTRL_XTT,KEY_CTRL_VARS,KEY_CTRL_CATALOG,KEY_CTRL_DEL,
   KEY_CHAR_EXPN,KEY_CHAR_LN,KEY_CHAR_LOG,KEY_CHAR_IMGNRY,',',KEY_CHAR_POW,
   KEY_CHAR_SIN,KEY_CHAR_COS,KEY_CHAR_TAN,KEY_CHAR_PI,KEY_CHAR_ROOT,KEY_CHAR_SQUARE,
   '7','8','9','(',')',-1,
   '4','5','6','*','/',-1,
   '1','2','3','+','-',-1,
   '0','.',KEY_CHAR_EXPN10,KEY_CHAR_ANS,KEY_CTRL_EXE,-1,
   // shifted
   KEY_SHIFT_LEFT,KEY_CTRL_PAGEUP,KEY_CTRL_PAGEDOWN,KEY_SHIFT_RIGHT,KEY_CTRL_OK,KEY_CTRL_EXIT,
   KEY_CTRL_MENU,KEY_PRGM_ACON,KEY_PRGM_ACON,9,10,11,
   KEY_CTRL_SHIFT,KEY_CTRL_ALPHA,KEY_SHIFT_ANS,KEY_CTRL_CLIP,KEY_CTRL_PASTE,KEY_CTRL_AC,
   KEY_CHAR_LBRCKT,KEY_CHAR_RBRCKT,KEY_CHAR_LBRACE,KEY_CHAR_RBRACE,'_',KEY_CHAR_STORE,
   KEY_CHAR_ASIN,KEY_CHAR_ACOS,KEY_CHAR_ATAN,'=','<','>',
   KEY_CTRL_F7,KEY_CTRL_F8,KEY_CTRL_F9,KEY_CTRL_F13,KEY_CTRL_F14,-1,
   KEY_CTRL_F4,KEY_CTRL_F5,KEY_CTRL_F6,KEY_CTRL_INS,'%',-1,
   KEY_CTRL_F1,KEY_CTRL_F2,KEY_CTRL_F3,'\t','\\',-1,
   KEY_CTRL_F10,KEY_CTRL_F11,KEY_CTRL_F12,KEY_SHIFT_ANS,KEY_CTRL_MENU,-1,
   // alpha
   KEY_CTRL_LEFT,KEY_CTRL_UP,KEY_CTRL_DOWN,KEY_CTRL_RIGHT,KEY_CTRL_OK,KEY_CTRL_EXIT,
   KEY_CTRL_MENU,KEY_PRGM_ACON,KEY_PRGM_ACON,9,10,11,
   KEY_CTRL_SHIFT,KEY_CTRL_ALPHA,':',';','"',KEY_CTRL_DEL,
   'a','b','c','d','e','f',
   'g','h','i','j','k','l',
   'm','n','o','p','q',-1,
   'r','s','t','u','v',-1,
   'w','x','y','z',' ',-1,
   '?','!',KEY_CHAR_EXPN10,KEY_CHAR_ANS,KEY_CTRL_EXE,-1,
   // alpha shifted
   KEY_SHIFT_LEFT,KEY_CTRL_PAGEUP,KEY_CTRL_PAGEDOWN,KEY_SHIFT_RIGHT,KEY_CTRL_OK,KEY_CTRL_EXIT,
   KEY_CTRL_MENU,KEY_PRGM_ACON,KEY_PRGM_ACON,9,10,11,
   KEY_CTRL_SHIFT,KEY_CTRL_ALPHA,':',';','\'','%',
   'A','B','C','D','E','F',
   'G','H','I','J','K','L',
   'M','N','O','P','Q',-1,
   'R','S','T','U','V',-1,
   'W','X','Y','Z',' ',-1,
   '?','!',KEY_CHAR_EXPN10,KEY_BOOK,KEY_CTRL_EXE,-1,
  };

int getkey(bool allow_suspend){
  int k=getkey_raw(allow_suspend);
  // translate
  return translated_keys[k];
}

// Casio prototype
void GetKey(int * key){
  *key=getkey(true);
}


void numworks_wait_1ms(int ms){
  for (int i=0;i<ms/128;++i){
#ifdef SIMULATOR
    Fl::wait(0.00001);
#endif
    Ion::Keyboard::State scan = Ion::Keyboard::scan();
      // if (scan!=16) std::cerr << scan << '\n';
    Ion::Keyboard::Key interruptKey = static_cast<Ion::Keyboard::Key>(Ion::Keyboard::Key::Back);
    if (scan.keyDown(interruptKey))
      return;
    Ion::Timing::msleep(128);
  }
  Ion::Timing::msleep(ms % 128);
}
